<?php

/**
 * Class DataCollector - Prepare data for import into WP posts
 *
 * @package Biuro Jobs importer
 * @used-by \jobs-importer\job-importer.php
 */
class DataCollector
{
    // Data mapping
    protected $dataMap = [
//        'lang'          => '@lang',
        'livas_id'      => '@livas_id',
        'valid'         => '@valid',
        'division'      => 'division',
        'name'          => 'title',
//        'position'      => 'PAREIGOS',
        'post_content'  => 'content',
        // 'tag'           => 'other_info',
        'type'          => 'type',
        'field'         => 'field',
        'city'          => 'city',
        'company'       => 'company',
        'email'         => 'email',
        'contact'       => 'contact',
        'post_date' => '@published_date',
        'post_modified' => '@updated',
    ];

    // Relations mapping
    protected $relationsMap = [
        'division'  =>  'division-id'
    ];

    // Fields, witch separates in metalanguages environment
    protected $multilingual = ['city', 'division', /*'position',*/ 'company', 'type', 'field'];

    // Records container's node tag name
    protected $containerTag = 'ads';

    // Item's node tag name
    protected $elementTag = 'ad';

    // Initial options for xml parser
    private $parserDefaults = [
        'namespaceSeparator' => ':',//you may want this to be something other than a colon
        'attributePrefix'    => '@',   //to distinguish between attributes and nodes with the same name
        'alwaysArray'        => array(),   //array of xml tag names which should always become arrays
        'autoArray'          => true,        //only create arrays for tags which appear more than once
        'textContent'        => '$',       //key used for the text content of elements
        'autoText'           => true,         //skip textContent key if node has no attributes or child nodes
        'keySearch'          => false,       //optional search and replace on tag and attribute names
        'keyReplace'         => false       //replace values for above search values (as passed to str_replace())
    ];

    // Source data
    protected $sourceData;

    protected $data;

    public function __construct($source)
    {
        $this->setData($source);
        $this->transformData();
    }

    /**
     *  Set data source:
     *  get data from source and transform into array
     *
     * @param $source
     */
    protected function setData($source)
    {
        $this->sourceData = $this->xmlToArray($this->loadData($source));
    }

    /**
     *  Parse xml data
     *
     * @param $source
     *
     * @return SimpleXMLElement
     */
    protected function loadData($source)
    {
        $context = stream_context_create([
            'http' => [
                'header' => 'Accept: application/xml',
            ],
            'ssl'  => [
                'verify_peer'      => false,
                'verify_peer_name' => false,
            ],
        ]);
        $xml = file_get_contents($source, false, $context);

        return simplexml_load_string($xml);
    }

    /**
     *  Get data
     *
     * @return mixed
     */
    public function getData()
    {
        return $this->data;
    }

    /**
     *  Transform data for import
     */
    protected function transformData()
    {
        $transformed = [];
        $parsedData = $this->sourceData;

        if (array_key_exists($this->containerTag, $parsedData)) {
            $container = $parsedData[$this->containerTag];
            if (array_key_exists($this->elementTag, $container)) {

                foreach ($container[$this->elementTag] as $elementData) {
                    var_dump($elementData);
                    $lang = $container[$this->elementTag]['@lang'];
                    $element = [];
                    foreach ($this->dataMap as $key => $item) {
                        $itemValue = (empty($elementData[$item])) ? '' : $elementData[$item];

                        var_dump($itemValue);
                        $value = (is_array($itemValue))
                            ? [$this->getKey($key, $itemValue, $lang) => html_entity_decode($this->getValue($itemValue))]
                            : html_entity_decode($itemValue);

                        $element[$key] = $value;

                        if (array_key_exists($key, $this->relationsMap)) {
                            $element['related_items'][$key] = $this->relationsMap[$key];
                        }
                    }

                    $transformed[] = $element;
                }
            }
        }

        var_dump($transformed);
        $this->data = $transformed;
    }

    /**
     *  Get value
     *
     * @param $value
     *
     * @return mixed|string
     */
    private function getValue($value)
    {
        if (is_array($value)) {
            if (array_key_exists('@key', $value)) {
                return (isset($value['$'])) ? $value['$'] : '';
            }
        }

        return $value;
    }

    /**
     *  Get key
     *
     * @param $key
     * @param $value
     * @param $lang
     *
     * @return string
     */
    private function getKey($key, $value, $lang)
    {
        if (is_array($value)) {
            if (array_key_exists('@key', $value)) {
                $key_id = '-' . $value['@key'];

                return $key . $key_id . $this->addLang($key, $lang);
            }
        }

        return $key . $this->addLang($key, $lang);
    }

    /**
     *  Adds language slug
     *
     * @param $key
     * @param $lang
     *
     * @return string
     */
    private function addLang($key, $lang)
    {
        if (in_array($key, $this->multilingual)) {
            return '-' . $lang;
        }

        return '';
    }


    /**
     *  XML parser, return array
     *
     * @param $xml
     * @param array $options
     *
     * @return array
     */
    function xmlToArray($xml, $options = array())
    {
        $options = array_merge($this->parserDefaults, $options);
        $namespaces = $xml->getDocNamespaces();
        $namespaces[''] = null; //add base (empty) namespace

        //get attributes from all namespaces
        $attributesArray = array();
        foreach ($namespaces as $prefix => $namespace) {
            foreach ($xml->attributes($namespace) as $attributeName => $attribute) {
                //replace characters in attribute name
                if ($options['keySearch']) {
                    $attributeName =
                        str_replace($options['keySearch'], $options['keyReplace'], $attributeName);
                }
                $attributeKey = $options['attributePrefix']
                    . ($prefix ? $prefix . $options['namespaceSeparator'] : '')
                    . $attributeName;
                $attributesArray[$attributeKey] = (string)$attribute;
            }
        }

        //get child nodes from all namespaces
        $tagsArray = array();
        foreach ($namespaces as $prefix => $namespace) {
            foreach ($xml->children($namespace) as $childXml) {
                //recurse into child nodes
                $childArray = $this->xmlToArray($childXml, $options);
//            list($childTagName, $childProperties) = each($childArray);
                foreach ($childArray as $key => $item) {
                    $childTagName = $key;
                    $childProperties = $item;
                }

                //replace characters in tag name
                if ($options['keySearch']) {
                    $childTagName =
                        str_replace($options['keySearch'], $options['keyReplace'], $childTagName);
                }
                //add namespace prefix, if any
                if ($prefix) {
                    $childTagName = $prefix . $options['namespaceSeparator'] . $childTagName;
                }

                if (!isset($tagsArray[$childTagName])) {
                    //only entry with this key
                    //test if tags of this type should always be arrays, no matter the element count
                    $tagsArray[$childTagName] =
                        in_array($childTagName, $options['alwaysArray']) || !$options['autoArray']
                            ? array($childProperties) : $childProperties;
                } elseif (
                    is_array($tagsArray[$childTagName]) && array_keys($tagsArray[$childTagName])
                    === range(0, count($tagsArray[$childTagName]) - 1)
                ) {
                    //key already exists and is integer indexed array
                    $tagsArray[$childTagName][] = $childProperties;
                } else {
                    //key exists so convert to integer indexed array with previous value in position 0
                    $tagsArray[$childTagName] = array($tagsArray[$childTagName], $childProperties);
                }
            }
        }

        //get text content of node
        $textContentArray = array();
        $plainText = trim((string)$xml);
        if ($plainText !== '') {
            $textContentArray[$options['textContent']] = $plainText;
        }

        //stick it all together
        $propertiesArray = !$options['autoText'] || $attributesArray || $tagsArray || ($plainText === '')
            ? array_merge($attributesArray, $tagsArray, $textContentArray) : $plainText;

        //return node as array
        return array(
            $xml->getName() => $propertiesArray
        );
    }

}
